import os
import sys

import click
import timeago
import tomli
from rich import box
from rich.table import Table

from .config import settings
from .console import console, cprint, eprint
from .operations import (
    SnapshotNotFound,
    create_snapshot,
    delete_snapshot,
    export_snapshot,
    find_snapshot,
    get_snapshots,
    import_snapshot,
    rename_snapshot,
    restore_snapshot,
)


def complete_snapshot_names(ctx, param, incomplete):
    """
    Returns a list of snapshot names for completion
    """
    db_url = os.environ.get("DATABASE_URL")
    if db_url:
        settings.initialize(url=db_url, debug=False)
        return [
            snapshot.name
            for snapshot in get_snapshots()
            if snapshot.name.startswith(incomplete)
        ]

    return []


def next_not_none(iterable):
    """
    Returns the next item in the iterable that is not None or ""
    """
    return next((item for item in iterable if item is not None and item != ""), None)


@click.group
@click.option(
    "--url",
    help="The database connection string to the database you want to take "
    "snapshots of."
    "\n\nExample: postgres://username:password@host:port/database_name",
)
@click.option("--debug/--no-debug", help="Show additional debugging information.")
def cli(url, debug):
    toml_params = {}
    try:
        with open("dslr.toml", "rb") as tomlf:
            toml_params = tomli.load(tomlf)
    except FileNotFoundError:
        pass

    config = {
        "url": next_not_none(
            [url, toml_params.get("url"), os.environ.get("DATABASE_URL")]
        )
        or "",
        "debug": next_not_none([debug, toml_params.get("debug"), False]),
    }

    # Update the settings singleton
    settings.initialize(**config)


@cli.command()
@click.argument("name", shell_complete=complete_snapshot_names)
@click.option(
    "-y",
    "--yes",
    "overwrite_confirmed",
    is_flag=True,
    help="Overwrite existing snapshot without confirmation.",
)
def snapshot(name: str, overwrite_confirmed: bool):
    """
    Takes a snapshot of the database
    """
    new = True

    try:
        snapshot = find_snapshot(name)

        if not overwrite_confirmed:
            click.confirm(
                click.style(
                    f"Snapshot {snapshot.name} already exists. Overwrite?", fg="yellow"
                ),
                abort=True,
            )

        delete_snapshot(snapshot)
        new = False
    except SnapshotNotFound:
        pass

    try:
        with console.status("Creating snapshot"):
            create_snapshot(name)
    except Exception as e:
        eprint("Failed to create snapshot")
        eprint(e, style="white")
        sys.exit(1)

    if new:
        cprint(f"Created new snapshot {name}", style="green")
    else:
        cprint(f"Updated snapshot {name}", style="green")


@cli.command()
@click.argument("name", shell_complete=complete_snapshot_names)
def restore(name):
    """
    Restores the database from a snapshot
    """
    try:
        snapshot = find_snapshot(name)
    except SnapshotNotFound:
        eprint(f"Snapshot {name} does not exist", style="red")
        sys.exit(1)

    with console.status("Restoring snapshot"):
        try:
            restore_snapshot(snapshot)
        except Exception as e:
            eprint("Failed to restore snapshot")
            eprint(e, style="white")
            sys.exit(1)

    cprint(f"Restored database from snapshot {snapshot.name}", style="green")


@cli.command()
def list():
    """
    Shows a list of all snapshots
    """
    try:
        snapshots = get_snapshots()
    except Exception as e:
        eprint("Failed to list snapshots")
        eprint(f"{e}", style="white")
        sys.exit(1)

    if len(snapshots) == 0:
        cprint("No snapshots found", style="yellow")
        return

    table = Table(box=box.SIMPLE)
    table.add_column("Name", style="cyan")
    table.add_column("Created")

    for snapshot in sorted(snapshots, key=lambda s: s.created_at, reverse=True):
        table.add_row(snapshot.name, timeago.format(snapshot.created_at))

    cprint(table)


@cli.command()
@click.argument("name", shell_complete=complete_snapshot_names)
def delete(name):
    """
    Deletes a snapshot
    """
    try:
        snapshot = find_snapshot(name)
    except SnapshotNotFound:
        eprint(f"Snapshot {name} does not exist", style="red")
        sys.exit(1)

    try:
        delete_snapshot(snapshot)
    except Exception as e:
        eprint("Failed to delete snapshot")
        eprint(e, style="white")
        sys.exit(1)

    cprint(f"Deleted snapshot {snapshot.name}", style="green")


@cli.command()
@click.argument("old_name", shell_complete=complete_snapshot_names)
@click.argument("new_name", shell_complete=complete_snapshot_names)
@click.option(
    "-y",
    "--yes",
    "overwrite_confirmed",
    is_flag=True,
    help="Overwrite existing snapshot without confirmation.",
)
def rename(old_name: str, new_name: str, overwrite_confirmed: bool):
    """
    Renames a snapshot
    """
    try:
        old_snapshot = find_snapshot(old_name)
    except SnapshotNotFound:
        eprint(f"Snapshot {old_name} does not exist", style="red")
        sys.exit(1)

    try:
        existing_snapshot = find_snapshot(new_name)

        if not overwrite_confirmed:
            click.confirm(
                click.style(
                    f"Snapshot {existing_snapshot.name} already exists. Overwrite?",
                    fg="yellow",
                ),
                abort=True,
            )

        delete_snapshot(existing_snapshot)
    except SnapshotNotFound:
        pass

    try:
        rename_snapshot(old_snapshot, new_name)
    except Exception as e:
        eprint("Failed to rename snapshot")
        eprint(e, style="white")
        sys.exit(1)

    cprint(f"Renamed snapshot {old_name} to {new_name}", style="green")


@cli.command()
@click.argument("name", shell_complete=complete_snapshot_names)
def export(name):
    """
    Exports a snapshot to a file
    """
    try:
        snapshot = find_snapshot(name)
    except SnapshotNotFound:
        eprint(f"Snapshot {name} does not exist", style="red")
        sys.exit(1)

    try:
        with console.status("Exporting snapshot"):
            export_path = export_snapshot(snapshot)
    except Exception as e:
        eprint("Failed to export snapshot")
        eprint(e, style="white")
        sys.exit(1)

    cprint(f"Exported snapshot {snapshot.name} to {export_path}", style="green")


@cli.command("import")
@click.argument("filename", type=click.Path(exists=True))
@click.argument("name", shell_complete=complete_snapshot_names)
@click.option(
    "-y",
    "--yes",
    "overwrite_confirmed",
    is_flag=True,
    help="Overwrite existing snapshot without confirmation.",
)
def import_(filename: str, name: str, overwrite_confirmed):
    """
    Imports a snapshot from a file
    """
    filename = click.format_filename(filename)

    try:
        snapshot = find_snapshot(name)

        if not overwrite_confirmed:
            click.confirm(
                click.style(
                    f"Snapshot {snapshot.name} already exists. Overwrite?", fg="yellow"
                ),
                abort=True,
            )

        delete_snapshot(snapshot)
    except SnapshotNotFound:
        pass

    try:
        with console.status("Importing snapshot"):
            import_snapshot(filename, name)
    except Exception as e:
        eprint("Failed to import snapshot")
        eprint(e, style="white")
        sys.exit(1)

    cprint(f"Imported snapshot {name} from {filename}", style="green")



from collections import namedtuple
from datetime import datetime
from typing import List, Optional

from psycopg2 import sql

from .config import settings
from .runner import exec_shell, exec_sql

################################################################################
# Database operations
################################################################################


def kill_connections(dbname: str):
    """
    Kills all connections to the given database
    """
    exec_sql(
        "SELECT pg_terminate_backend(pg_stat_activity.pid) FROM pg_stat_activity "
        "WHERE pg_stat_activity.datname = %s",
        [dbname],
    )


def create_database(*, dbname: str, template: Optional[str] = None):
    """
    Creates a new database with the given name, optionally using the given template
    """
    if template:
        exec_sql(
            sql.SQL("CREATE DATABASE {} TEMPLATE {}").format(
                sql.Identifier(dbname),
                sql.Identifier(template),
            )
        )
    else:
        exec_sql(
            sql.SQL("CREATE DATABASE {}").format(
                sql.Identifier(dbname),
            )
        )


def drop_database(dbname: str):
    """
    Drops the given database
    """
    exec_sql(sql.SQL("DROP DATABASE {}").format(sql.Identifier(dbname)))


################################################################################
# Snapshot operations
################################################################################

Snapshot = namedtuple("Snapshot", ["dbname", "name", "created_at"])


def generate_snapshot_db_name(
    snapshot_name: str, created_at: Optional[datetime] = None
) -> str:
    """
    Generates a database name for a snapshot
    """
    if not created_at:
        created_at = datetime.now()

    timestamp = round(created_at.timestamp())

    return f"dslr_{timestamp}_{snapshot_name}"


def get_snapshots() -> List[Snapshot]:
    """
    Returns the list of database snapshots

    Snapshots are databases that follow the naming convention:

    dslr_<timestamp>_<snapshot_name>
    """
    # Find the snapshot databases
    result = exec_sql("SELECT datname FROM pg_database WHERE datname LIKE 'dslr_%'")

    if result is None:
        raise RuntimeError("Did not get results from database.")

    snapshot_dbnames = sorted([row[0] for row in result])

    # Parse the name into a Snapshot
    parts = [dbname.split("_") for dbname in snapshot_dbnames]
    return [
        Snapshot(
            dbname=line,
            name="_".join(part[2:]),
            created_at=datetime.fromtimestamp(int(part[1])),
        )
        for part, line in zip(parts, snapshot_dbnames)
    ]


class SnapshotNotFound(Exception):
    pass


def find_snapshot(snapshot_name: str) -> Snapshot:
    """
    Returns the snapshot with the given name
    """
    snapshots = get_snapshots()

    try:
        return next(
            snapshot for snapshot in snapshots if snapshot.name == snapshot_name
        )
    except StopIteration as e:
        raise SnapshotNotFound(
            f'Snapshot with name "{snapshot_name}" does not exist.'
        ) from e


def create_snapshot(snapshot_name: str):
    """
    Takes a snapshot of the database

    Snapshotting works by creating a new database using the local database as a
    template.
    """
    kill_connections(settings.db.name)
    create_database(
        dbname=generate_snapshot_db_name(snapshot_name), template=settings.db.name
    )


def delete_snapshot(snapshot: Snapshot):
    """
    Deletes the given snapshot
    """
    drop_database(snapshot.dbname)


def restore_snapshot(snapshot: Snapshot):
    """
    Restores the database from the given snapshot
    """
    kill_connections(settings.db.name)
    drop_database(settings.db.name)
    create_database(dbname=settings.db.name, template=snapshot.dbname)


def rename_snapshot(snapshot: Snapshot, new_name: str):
    """
    Renames the given snapshot
    """
    exec_sql(
        sql.SQL("ALTER DATABASE {} RENAME TO {}").format(
            sql.Identifier(snapshot.dbname),
            sql.Identifier(generate_snapshot_db_name(new_name, snapshot.created_at)),
        )
    )


def export_snapshot(snapshot: Snapshot) -> str:
    """
    Exports the given snapshot to a file
    """
    export_path = f"{snapshot.name}_{snapshot.created_at:%Y%m%d-%H%M%S}.dump"
    exec_shell("pg_dump", "-Fc", "-d", snapshot.dbname, "-f", export_path)

    return export_path


def import_snapshot(import_path: str, snapshot_name: str):
    """
    Imports the given snapshot from a file
    """
    dbname = generate_snapshot_db_name(snapshot_name)
    create_database(dbname=dbname)

    exec_shell("pg_restore", "-d", dbname, "--no-acl", "--no-owner", import_path)



import os
import subprocess
from collections import namedtuple
from typing import Any, List, Optional, Tuple, Union

from psycopg2 import sql

from dslr.pg_client import PGClient

from .config import settings
from .console import console

Result = namedtuple("Result", ["stdout", "stderr"])


def exec_shell(*cmd: str) -> Result:
    """
    Executes a command.
    """

    # Set PG environment variables based on the settings
    env = os.environ.copy()
    env["PGHOST"] = settings.db.host or env.get("PGHOST", "")
    env["PGPORT"] = str(settings.db.port) or env.get("PGPORT", "")
    env["PGUSER"] = settings.db.username or env.get("PGUSER", "")
    env["PGPASSWORD"] = settings.db.password or env.get("PGPASSWORD", "")

    if settings.debug:
        console.log(f"COMMAND: {cmd}")

    with subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=env,
    ) as p:
        stdout, stderr = p.communicate()

        if settings.debug:
            console.log("STDOUT:\n", stdout.decode("utf-8"), "\n")
            console.log("STDERR:\n", stderr.decode("utf-8"), "\n")

        if p.returncode != 0:
            raise RuntimeError(f"Command failed: {cmd}")

        return Result(
            stdout=stdout.decode("utf-8"),
            stderr=stderr.decode("utf-8"),
        )


# Singleton instance of PGClient
pg_client: Optional[PGClient] = None


def exec_sql(
    sql: Union[sql.Composed, str], data: Optional[List[Any]] = None
) -> Optional[List[Tuple[Any, ...]]]:
    """
    Executes a SQL query.
    """
    global pg_client

    if not pg_client:
        # We always want to connect to the `postgres` and not the target
        # database because none of our operations need to query the target
        # database.
        pg_client = PGClient(
            host=settings.db.host,
            port=settings.db.port,
            user=settings.db.username,
            password=settings.db.password,
            dbname="postgres",
        )

    return pg_client.execute(sql, data)



from typing import Any, List, Optional, Tuple

import psycopg2
import psycopg2.extensions

from dslr.console import console

from .config import settings


class PGClient:
    """
    Thin wrapper around psycopg2
    """

    def __init__(self, host, port, user, password, dbname):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.dbname = dbname

        self.conn = psycopg2.connect(
            host=host,
            port=port,
            user=user,
            password=password,
            dbname=dbname,
        )
        self.conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)

        self.cur = self.conn.cursor()

    def execute(self, sql, data) -> Optional[List[Tuple[Any, ...]]]:
        if settings.debug:
            console.log(f"SQL: {sql}")
            console.log(f"DATA: {data}")

        self.cur.execute(sql, data)

        try:
            result = self.cur.fetchall()
        except psycopg2.ProgrammingError:
            result = None

        return result



from functools import partial

from rich.console import Console

console = Console()
error_console = Console(stderr=True)

cprint = partial(console.print, highlight=False)
eprint = partial(error_console.print, style="bold red", highlight=False)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dslr']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'rich>=12.5.1,<13.0.0',
 'timeago>=1.0.15,<2.0.0',
 'tomli>=2.0.1,<3.0.0']

entry_points = \
{'console_scripts': ['dslr = dslr.cli:cli']}

setup_kwargs = {
    'name': 'dslr',
    'version': '0.3.1a0',
    'description': 'Take lightning fast snapshots of your local Postgres databases.',
    'long_description': '<br />\n<br />\n<p align="center">\n  <a href="https://github.com/mixxorz/DSLR">\n    <img width="281" height="84" src="https://user-images.githubusercontent.com/3102758/181914025-44bff27e-aac1-4d1b-a037-9fa98f9fed65.png" alt="The DSLR logo">\n  </a>\n</p>\n<br />\n\n<p align="center">\n  <a href=""><img src="" alt=""></a>\n  <a href="https://badge.fury.io/py/dslr"><img src="https://badge.fury.io/py/dslr.svg" alt="PyPI version"></a>\n  <a href="https://pypi.python.org/pypi/dslr/"><img src="https://img.shields.io/pypi/pyversions/dslr.svg" alt="PyPI Supported Python Versions"></a>\n  <a href="https://github.com/mixxorz/dslr"><img src="https://github.com/mixxorz/dslr/actions/workflows/tests.yml/badge.svg" alt="GitHub Actions (Code quality and tests)"></a>\n\n</p>\n\n<p align="center">\n  <img src="https://user-images.githubusercontent.com/3102758/183229767-4501e6aa-e1cf-43c7-bd55-61faaa249ca2.png" alt="A terminal showing DSLR\'s command line interface.">\n</p>\n\n---\n\nDatabase Snapshot, List, and Restore\n\nTake lightning fast snapshots of your local Postgres databases.\n\n## What is this?\n\nDSLR is a tool that allows you to quickly take and restore database snapshots\nwhen you\'re writing database migrations, switching branches, or messing with\nSQL.\n\nIt\'s meant to be a spiritual successor to\n[Stellar](https://github.com/fastmonkeys/stellar).\n\n**Important:** DSLR is intended for development use only. It is not advisable to\nuse DSLR on production databases.\n\n## Performance\n\nDSLR is much faster than the standard `pg_dump`/`pg_restore` approach to snapshots.\n\n<p align="center">\n  <img src="https://user-images.githubusercontent.com/3102758/182014327-1b13da6e-63ad-4bbe-817e-7d6c66befc98.png" alt="A chart comparing the execution time between DSLR and pg_dump/pg_restore. For snapshot and restore, DSLR took 4.125 seconds and 4.431 seconds respectively. pg_dump/pg_restore took 36.602 seconds and 13.257 seconds respectively.">\n</p>\n\nDSLR is 8x faster at taking snapshots and 3x faster at restoring snapshots compared to the `pg_dump`/`pg_restore` approach.\n\n<details>\n  <summary>Testing methodology</summary>\n  \n  I spun up Postgres 12.3 using Docker, created a test database, and filled it with 1GB of random data using this script:\n  \n  ```SQL\n  CREATE TABLE large_test (num1 bigint, num2 double precision, num3 double precision);\n\nINSERT INTO large*test (num1, num2, num3)\nSELECT round(random() * 10), random(), random() \\_ 142\nFROM generate_series(1, 20000000) s(i);\n\n```\n\nI used the following commands to measure the execution time:\n\n```\n\ntime dslr snapshot my-snapshot\ntime dslr restore my-snapshot\ntime pg_dump -Fc -f export.dump\ntime pg_restore --no-acl --no-owner export.dump\n\n```\n\nI ran each command three times and plotted the mean in the chart.\n\nHere\'s the raw data:\n\n| Command       | Run | Execution time (seconds) |\n| ------------- | --- | ------------------------ |\n| dslr snapshot | 1   | 4.797                    |\n|               | 2   | 4.650                    |\n|               | 3   | 2.927                    |\n| dslr restore  | 1   | 5.840                    |\n|               | 2   | 4.122                    |\n|               | 3   | 3.331                    |\n| pg_dump       | 1   | 37.345                   |\n|               | 2   | 36.227                   |\n|               | 3   | 36.233                   |\n| pg_restore    | 1   | 13.304                   |\n|               | 2   | 13.148                   |\n|               | 3   | 13.320                   |\n</details>\n\n## Install\n\n```\n\npip install DSLR psycopg2 # or psycopg2-binary\n\n````\n\nAdditionally, the DSLR `export` and `import` snapshot commands require `pg_dump`\nand `pg_restore` to be present in your `PATH`.\n\n## Configuration\n\nYou can tell DSLR which database to take snapshots of in a few ways:\n\n**DATABASE_URL**\n\nIf the `DATABASE_URL` environment variable is set, DSLR will use this to connect\nto your target database.\n\n```bash\nexport DATABASE_URL=postgres://username:password@host:port/database_name\n````\n\n**dslr.toml**\n\nIf a `dslr.toml` file exists in the current directory, DSLR will read its\nsettings from there. DSLR will prefer this over the environment variable.\n\n```toml\nurl = \'postgres://username:password@host:port/database_name\'\n```\n\n**`--url` option**\n\nFinally, you can explicitly pass the connection string via the `--url` option.\nThis will override any of the above settings.\n\n## Usage\n\n```\n$ dslr snapshot my-first-snapshot\nCreated new snapshot my-first-snapshot\n\n$ dslr restore my-first-snapshot\nRestored database from snapshot my-first-snapshot\n\n$ dslr list\n\n  Name                Created\n ────────────────────────────────────\n  my-first-snapshot   2 minutes ago\n\n$ dslr rename my-first-snapshot fresh-db\nRenamed snapshot my-first-snapshot to fresh-db\n\n$ dslr delete some-old-snapshot\nDeleted some-old-snapshot\n\n$ dslr export my-feature-test\nExported snapshot my-feature-test to my-feature-test_20220730-075650.dump\n\n$ dslr import snapshot-from-a-friend_20220730-080632.dump friend-snapshot\nImported snapshot friend-snapshot from snapshot-from-a-friend_20220730-080632.dump\n```\n\n## How does it work?\n\nDSLR takes snapshots by cloning databases using Postgres\' [Template\nDatabases](https://www.postgresql.org/docs/current/manage-ag-templatedbs.html)\nfunctionality. This is the main source of DSLR\'s speed.\n\nThis means that taking a snapshot is just creating a new database using the main\ndatabase as the template. Restoring a snapshot is just deleting the main\ndatabase and creating a new database using the snapshot database as the\ntemplate. So on and so forth.\n\n## License\n\nMIT\n',
    'author': 'Mitchel Cabuloy',
    'author_email': 'mixxorz@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mixxorz/DSLR',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from urllib.parse import urlparse

from .console import console


@dataclass
class DatabaseConnection:
    host: str
    port: int
    username: str
    password: str
    name: str


class Settings:
    url: str
    debug: bool

    db: DatabaseConnection

    def initialize(self, *, url: str, debug: bool):
        self.url = url
        self.debug = debug

        parsed = urlparse(url)

        self.db = DatabaseConnection(
            host=parsed.hostname or "",
            port=parsed.port or 5432,
            username=parsed.username or "",
            password=parsed.password or "",
            name=parsed.path[1:],
        )

        if debug:
            console.log(f"URL: {self.url}")
            console.log(f"DB: {self.db}")


# Settings singleton
settings = Settings()
